import { EntityComponentTypes, EquipmentSlot, ItemStack, world } from "@minecraft/server";
import { Dimension, herobrineData, herobrineEvents, herobrineSettings } from "./index.js";

export const updateHerobrineData = () => { writeDynamicData("herobrineData", herobrineData) }
export const updateDynamicSettings = () => { writeDynamicData("herobrineSettings", herobrineSettings) }
export const updateDynamicEvents = () => { writeDynamicData("herobrineEvents", herobrineEvents) }

export const findIndexPerString = (list, input = "test") => list.findIndex(obj => { return input.toLocaleLowerCase() == obj.name.toLocaleLowerCase() })

export function isAllowedBlock(block, suffixList = allowedSuffixes_list, tagList = allowedTags_list) {
    if (!block) return false;

    // Se o bloco possui a propriedade "flower_type", é permitido
    if (block?.permutation?.getAllStates?.()["flower_type"]) {
        return true;
    }

    // Se o bloco possui alguma das tags permitidas
    if (tagList.some(tag => block.hasTag(tag))) {
        return true;
    }

    // Verifica se o block.typeId termina com algum dos sufixos permitidos
    const typeId = block.typeId;
    for (const suffix of suffixList) {
        if (typeId.endsWith(suffix)) {
            return true;
        }
    }

    return false;
}

export function equipRandomItemsToZombie(zombie) {
    const swordsList = ["minecraft:iron_sword", "minecraft:stone_sword", "minecraft:golden_sword", "minecraft:wooden_sword"];

    const helmetList = ["minecraft:iron_helmet", "minecraft:golden_helmet", "minecraft:chainmail_helmet"];
    const chestplateList = ["minecraft:iron_chestplate", "minecraft:golden_chestplate", "minecraft:chainmail_chestplate"];
    const leggingsList = ["minecraft:iron_leggings", "minecraft:golden_leggings", "minecraft:chainmail_leggings"];
    const bootsList = ["minecraft:iron_boots", "minecraft:golden_boots", "minecraft:chainmail_boots"];

    zombie.runCommand(`replaceitem entity @s slot.weapon.mainhand 0 ${swordsList[randomNumber(0, swordsList.length - 1)]}`)

    zombie.runCommand(`replaceitem entity @s slot.armor.head 0 ${helmetList[randomNumber(0, helmetList.length - 1)]}`)
    zombie.runCommand(`replaceitem entity @s slot.armor.chest 0 ${chestplateList[randomNumber(0, chestplateList.length - 1)]}`)
    zombie.runCommand(`replaceitem entity @s slot.armor.legs 0 ${leggingsList[randomNumber(0, leggingsList.length - 1)]}`)
    zombie.runCommand(`replaceitem entity @s slot.armor.feet 0 ${bootsList[randomNumber(0, bootsList.length - 1)]}`)

    zombie.addEffect("fire_resistance", 20 * 60, { showParticles: false })
}

export function calculateDistance(vetor1, vetor2) {
    const dx = vetor2.x - vetor1.x;
    const dy = vetor2.y - vetor1.y;
    const dz = vetor2.z - vetor1.z;
    return Math.sqrt(dx * dx + dy * dy + dz * dz).toFixed(3);
}

export function herobrineVFX(location, id = "minecraft:ice_evaporation_emitter") {
    const step = 0.1;
    const minHeight = -0.6;
    const maxHeight = 0.80;
    const radius = 0.65;
    const particles = 5;

    for (let height = minHeight; height < maxHeight; height += step) {
        VFX(location, id, radius, particles, height);
    }
}

export function VFX(location, particleID, radius, maxParticles = 100, yFIX = 0) {
    // Função auxiliar para limitar o valor entre um mínimo e máximo
    const clamp = (value, min, max) => Math.max(min, Math.min(value, max));

    // Extrai as coordenadas do local central
    const { x: centerX, y: centerY, z: centerZ } = location;

    // Calcula a quantidade de partículas (garante que seja um número inteiro)
    const particleCount = Math.min(maxParticles, Math.floor(360 / radius));

    for (let i = 0; i < particleCount; i++) {
        // Gera ângulos aleatórios para coordenadas esféricas
        const theta = Math.random() * 2 * Math.PI;       // Ângulo azimutal (0 a 2PI)
        const phi = Math.acos(2 * Math.random() - 1);      // Ângulo polar (0 a PI)

        // Define um raio aleatório dentro da esfera com fator de variação
        const distance = radius * Math.random() * 0.5;

        // Converte as coordenadas esféricas para cartesianas
        const particleX = centerX + distance * Math.sin(phi) * Math.cos(theta) - 0.5;
        const particleY = clamp(centerY + distance * Math.cos(phi) + yFIX, -64, 320);
        const particleZ = centerZ + distance * Math.sin(phi) * Math.sin(theta) - 0.5;

        // Spawna a partícula no mundo
        Dimension.spawnParticle(particleID, { x: particleX, y: particleY, z: particleZ });
    }
}

export function randomNumber(min, max, round = true) {
    if (min > max) return null;

    let rnd;
    if (typeof crypto !== "undefined" && crypto.getRandomValues) {
        const buffer = new Uint32Array(1);
        crypto.getRandomValues(buffer);
        rnd = buffer[0] / 4294967296; // Valor entre 0 (incluso) e 1 (excluso)
    } else {
        rnd = Math.random();
    }

    const value = rnd * (max - min) + min;
    return round ? Math.round(value) : value;
}

export function fetchDynamicData(name) {
    try {
        return JSON.parse(world.getDynamicProperty(name))
    } catch {
        console.warn("Erro na conversão, pode ser que esse Dynamic não exista.", name)
        return null
    }
}

export function writeDynamicData(name, input = [{ none: true }]) {
    const inputFormat = JSON.stringify(input)
    world.setDynamicProperty(name, inputFormat)
}

export function updateDynamicData(dynamicName, structureConfig = null) {
    try {

        const dynamicData = fetchDynamicData(dynamicName);

        if (dynamicData == null || dynamicData == undefined) {
            writeDynamicData(dynamicName, structureConfig);
        }

        return dynamicData

    } catch (error) {
        writeDynamicData(dynamicName, structureConfig);
        return fetchDynamicData(dynamicName);
    }
}

export async function spawnEntityInRegion(entityID, loc, rad = 4, height = 1, minDistance = 5) {
    try {
        const { x, y, z } = loc;
        const [startX, endX] = [x - rad, x + rad];
        const [startZ, endZ] = [z - rad, z + rad];
        const baseY = Math.ceil(y);
        const [startY, endY] = [baseY - height - 1, baseY + height];
        let chosenPosition = null, validCount = 0;

        for (let dy = startY; dy < endY; dy++) {
            for (let dx = startX; dx <= endX; dx++) {
                const absDiffX = Math.abs(dx - x);
                for (let dz = startZ; dz <= endZ; dz++) {
                    const absDiffZ = Math.abs(dz - z);
                    if (absDiffX <= minDistance && absDiffZ <= minDistance) continue;

                    const blockAbove = Dimension.getBlock({ x: dx, y: dy + 1, z: dz });
                    const blockCurrent = Dimension.getBlock({ x: dx, y: dy, z: dz });
                    const blockFloor = Dimension.getBlock({ x: dx, y: dy - 1, z: dz });

                    if (!blockAbove.isAir || !blockCurrent.isAir || blockFloor.isAir || blockFloor.isLiquid || isAllowedBlock(blockFloor, allowedSuffixesHerobrine_list, allowedTags_list)) continue;

                    validCount++;
                    if (Math.random() < 1 / validCount) chosenPosition = { x: dx, y: dy, z: dz };
                }
            }
        }

        return chosenPosition ? Dimension.spawnEntity(entityID, chosenPosition) : null;
    } catch (error) {
        console.error("Erro ao executar spawnEntityInRegion:", error);
        return null;
    }
}

export class BlockManager {

    replaceBlock(vector3, blockInLoc, replacementBlock, keepPermutation) {
        if (keepPermutation) {
            const permutationData = blockInLoc.permutation.getState("torch_facing_direction");
            Dimension.setBlockType(vector3, replacementBlock);
            Dimension.setBlockPermutation(
                vector3,
                blockInLoc.permutation.withState("torch_facing_direction", permutationData)
            );
        } else {
            Dimension.setBlockType(vector3, replacementBlock);
        }
    }

    squareGetBlocks({
        locationStart,
        allSides = false,
        blockTargetID = "none",
        replacementBlock = "minecraft:redstone_torch",
        keepPermutation = true,
        rad = 30,
        height = 5
    } = {}) {
        try {
            const { x: originX, y: originY, z: originZ } = locationStart;

            const [minX, minY, minZ] = [
                Math.round(originX - rad),
                Math.round(allSides ? originY - height : originY),
                Math.round(originZ - rad)
            ];
            const [maxX, maxY, maxZ] = [
                Math.round(originX + rad),
                Math.round(originY + height),
                Math.round(originZ + rad)
            ];

            for (let x = minX; x <= maxX; x++) {
                for (let y = minY; y <= maxY; y++) {
                    for (let z = minZ; z <= maxZ; z++) {
                        const vector3 = { x: x, y: y, z: z };
                        const blockInLoc = Dimension.getBlock(vector3);

                        if (blockTargetID !== "none" && blockInLoc.typeId === blockTargetID && blockInLoc.typeId !== "minecraft:air") {
                            this.replaceBlock(vector3, blockInLoc, replacementBlock, keepPermutation);
                        }
                    }
                }
            }
        } catch { }
    }
};

const allowedSuffixesHerobrine_list = [
    "_double_slab",
    "_fence_gate",
    "_carpet",
    "_pressure_plate",
    "_button",
    "_wire",
    "_repeater",
    "_comparator",
    "_banner",
    "_sign",
    "_leaves",
    "_rod",
    "minecraft:lever",
    "minecraft:torch",
    "minecraft:bell",
    "minecraft:tripwire_hook",
    "minecraft:rose_bush",
    "minecraft:lilac",
    "minecraft:peony",
    "minecraft:sunflower",
    "minecraft:heavy_core",
    "minecraft:web",
    "minecraft:bed",
    "minecraft:daylight_detector",
    "minecraft:deadbush",
    "minecraft:pointed_dripstone",
    "minecraft:daylight_detector_inverted",
    "minecraft:short_grass",
    "minecraft:tall_grass",
    "minecraft:bamboo",
    "minecraft:fern",
    "minecraft:cocoa",
    "minecraft:vine",
    "minecraft:dandelion",
    "minecraft:fire",
];

export const allowedSuffixes_list = [
    "_slab",
    "_double_slab",
    "_fence_gate",
    "_carpet",
    "_pressure_plate",
    "_button",
    "_wire",
    "_repeater",
    "_comparator",
    "_banner",
    "_layer",
    "_sign",
    "_leaves",
    "_rod",
    "minecraft:lever",
    "minecraft:torch",
    "minecraft:bell",
    "minecraft:tripwire_hook",
    "minecraft:rose_bush",
    "minecraft:lilac",
    "minecraft:peony",
    "minecraft:sunflower",
    "minecraft:heavy_core",
    "minecraft:web",
    "minecraft:bed",
    "minecraft:daylight_detector",
    "minecraft:deadbush",
    "minecraft:pointed_dripstone",
    "minecraft:daylight_detector_inverted",
    "minecraft:short_grass",
    "minecraft:tall_grass",
    "minecraft:bamboo",
    "minecraft:fern",
    "minecraft:cocoa",
    "minecraft:vine",
    "minecraft:dandelion",
    "minecraft:fire",
];

export const allowedTags_list = [
    "minecraft:not_feature_replaceable",
    "minecraft:plant",
    "minecraft:rail",
    "minecraft:trapdoors",
    "minecraft:snow",
    "minecraft:text_sing",
];